



echo "%CMAKE_GENERATOR%"
IF %ERRORLEVEL% NEQ 0 exit /B 1
echo "%CMAKE_GENERATOR_PLATFORM%"
IF %ERRORLEVEL% NEQ 0 exit /B 1
echo "%CMAKE_GENERATOR_TOOLSET%"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
